import logging
from typing import List

from peek_abstract_chunked_index.private.server.controller.ACIProcessorQueueControllerABC import \
    ACIProcessorQueueControllerABC, ACIProcessorQueueBlockItem
from peek_abstract_chunked_index.private.server.controller.ACIProcessorStatusNotifierABC import \
    ACIProcessorStatusNotifierABC
from peek_abstract_chunked_index.private.tuples.ACIProcessorQueueTupleABC import \
    ACIProcessorQueueTupleABC
from peek_plugin_livedb._private.server.controller.AdminStatusController import \
    AdminStatusController
from peek_plugin_livedb._private.storage.LiveDbItem import LiveDbItem
from peek_plugin_livedb._private.storage.LiveDbModelSet import getOrCreateLiveDbModelSet
from peek_plugin_livedb._private.storage.LiveDbRawValueQueue import LiveDbRawValueQueue
from peek_plugin_livedb.tuples.LiveDbRawValueUpdateTuple import LiveDbRawValueUpdateTuple
from vortex.DeferUtil import deferToThreadWrapWithLogger

logger = logging.getLogger(__name__)


class _Notifier(ACIProcessorStatusNotifierABC):
    def __init__(self, adminStatusController: AdminStatusController):
        self._adminStatusController = adminStatusController

    def setProcessorStatus(self, state: bool, queueSize: int):
        self._adminStatusController.status.rawValueQueueStatus = state
        self._adminStatusController.status.rawValueQueueSize = queueSize
        self._adminStatusController.notify()

    def addToProcessorTotal(self, delta: int):
        self._adminStatusController.status.rawValueProcessedTotal += delta
        self._adminStatusController.notify()

    def setProcessorError(self, error: str):
        self._adminStatusController.status.rawValueLastError = error
        self._adminStatusController.notify()


class LiveDbValueUpdateQueueController(ACIProcessorQueueControllerABC):
    # Prioritize the livedb updater.
    MAX_CPU_PERCENTAGE = 85.00

    QUEUE_ITEMS_PER_TASK = 500
    POLL_PERIOD_SECONDS = 0.200

    QUEUE_BLOCKS_MAX = 20
    QUEUE_BLOCKS_MIN = 4

    WORKER_TASK_TIMEOUT = 60.0

    _logger = logger
    _QueueDeclarative: ACIProcessorQueueTupleABC = LiveDbRawValueQueue
    _VacuumDeclaratives = (LiveDbRawValueQueue, LiveDbItem)

    def __init__(self, ormSessionCreator, adminStatusController: AdminStatusController):
        ACIProcessorQueueControllerABC.__init__(self, ormSessionCreator,
                                                _Notifier(adminStatusController))

        self._modelSetIdByKey = {}

    def _sendToWorker(self, block: ACIProcessorQueueBlockItem):
        from peek_plugin_livedb._private.worker.tasks.LiveDbItemUpdateTask import \
            updateValues

        return updateValues.delay(block.itemsEncodedPayload)

    def _processWorkerResults(self, results):
        pass

    # ---------------
    # Deduplicate method

    def _dedupeQueueSql(self, lastFetchedId: int, dedupeLimit: int):
        """ Deduplicate Queue

        NOTE: In this SQL, we take the last value added to the queue, as it will have
        the latest value.

        """
        return '''
                 with sq_raw as (
                    SELECT "id", "modelSetId", "key"
                    FROM pl_livedb."LiveDbRawValueQueue"
                    WHERE id > %(id)s
                    LIMIT %(limit)s
                ), sq as (
                    -- Select the latest queued update, this does require that updates
                    -- come in, in order
                    SELECT max(id) as "maxId", "modelSetId", "key"
                    FROM sq_raw
                    GROUP BY  "modelSetId", "key"
                    HAVING count("key") > 1
                )
                DELETE
                FROM pl_livedb."LiveDbRawValueQueue"
                     USING sq sq1
                WHERE pl_livedb."LiveDbRawValueQueue"."id" != sq1."maxId"
                    AND pl_livedb."LiveDbRawValueQueue"."id" > %(id)s
                    AND pl_livedb."LiveDbRawValueQueue"."modelSetId" = sq1."modelSetId"
                    AND pl_livedb."LiveDbRawValueQueue"."key" = sq1."key"
                    
            ''' % {'id': lastFetchedId, 'limit': dedupeLimit}

    # ---------------
    # Insert into Queue methods

    @deferToThreadWrapWithLogger(logger)
    def queueData(self, modelSetKey: str,
                  updates: List[LiveDbRawValueUpdateTuple]):
        if not updates:
            return

        ormSession = self._dbSessionCreator()
        try:
            logger.debug("Queueing %s raw values for compile", len(updates))

            if modelSetKey not in self._modelSetIdByKey:
                modelSet = getOrCreateLiveDbModelSet(ormSession, modelSetKey=modelSetKey)
                self._modelSetIdByKey[modelSet.key] = modelSet.id

            modelSetId = self._modelSetIdByKey[modelSetKey]

            inserts = []
            for update in updates:
                inserts.append(dict(modelSetId=modelSetId,
                                    key=update.key,
                                    rawValue=update.rawValue))

            ormSession.execute(LiveDbRawValueQueue.__table__.insert(), inserts)
            ormSession.commit()

        finally:
            ormSession.close()



"""added queue table

Peek Plugin Database Migration Script

Revision ID: d4ffb9c5cb27
Revises: 19b2a8d326ad
Create Date: 2020-04-18 16:15:31.166637

"""

# revision identifiers, used by Alembic.
revision = 'd4ffb9c5cb27'
down_revision = '19b2a8d326ad'
branch_labels = None
depends_on = None

import sqlalchemy as sa
from alembic import op


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('LiveDbRawValueQueue',
                    sa.Column('id', sa.BigInteger(), autoincrement=True, nullable=False),
                    sa.Column('modelSetId', sa.Integer(), nullable=False),
                    sa.Column('key', sa.String(), nullable=False),
                    sa.Column('rawValue', sa.String(), nullable=False),
                    sa.PrimaryKeyConstraint('id', 'modelSetId', 'key', 'rawValue'),
                    schema='pl_livedb'
                    )

    op.add_column('LiveDbModelSet', sa.Column('key', sa.String(), nullable=True),
                  schema='pl_livedb')

    op.execute('''UPDATE "pl_livedb"."LiveDbModelSet"
                  SET "key" = "name"''')

    op.alter_column('LiveDbModelSet', 'key', type_=sa.String(), nullable=False,
                    schema='pl_livedb')
    op.create_unique_constraint('idx_liveDbModelSet_key', 'LiveDbModelSet', ['key'],
                                schema='pl_livedb')

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint('idx_liveDbModelSet_key', 'LiveDbModelSet', schema='pl_livedb',
                       type_='unique')
    op.drop_column('LiveDbModelSet', 'key', schema='pl_livedb')
    op.drop_table('LiveDbRawValueQueue', schema='pl_livedb')
    # ### end Alembic commands ###



import logging
from datetime import datetime
from typing import List

import pytz
from sqlalchemy.sql.expression import select
from txcelery.defer import DeferrableTask

from peek_plugin_base.storage.StorageUtil import makeCoreValuesSubqueryCondition
from peek_plugin_base.worker import CeleryDbConn
from peek_plugin_livedb._private.storage.LiveDbItem import LiveDbItem
from peek_plugin_livedb._private.storage.LiveDbModelSet import getOrCreateLiveDbModelSet
from peek_plugin_base.worker.CeleryApp import celeryApp
from peek_plugin_livedb.tuples.ImportLiveDbItemTuple import ImportLiveDbItemTuple

logger = logging.getLogger(__name__)


@DeferrableTask
@celeryApp.task(bind=True)
def importLiveDbItems(self, modelSetKey: str,
                      newItems: List[ImportLiveDbItemTuple]) -> List[str]:
    """ Compile Grids Task

    :param self: A celery reference to this task
    :param modelSetKey: The model set name
    :param newItems: The list of new items
    :returns: A list of grid keys that have been updated.
    """

    startTime = datetime.now(pytz.utc)

    session = CeleryDbConn.getDbSession()
    engine = CeleryDbConn.getDbEngine()
    conn = engine.connect()
    transaction = conn.begin()

    liveDbTable = LiveDbItem.__table__
    try:

        liveDbModelSet = getOrCreateLiveDbModelSet(session, modelSetKey)

        # This will remove duplicates
        itemsByKey = {i.key: i for i in newItems}

        allKeys = list(itemsByKey)
        existingKeys = set()

        # Query for existing keys, in 1000 chinks
        chunkSize = 1000
        offset = 0
        while True:
            chunk = allKeys[offset:offset + chunkSize]
            if not chunk:
                break
            offset += chunkSize
            stmt = (select([liveDbTable.c.key])
                    .where(liveDbTable.c.modelSetId == liveDbModelSet.id)
            .where(makeCoreValuesSubqueryCondition(
                engine, liveDbTable.c.key, chunk
            ))
            )

            result = conn.execute(stmt)

            existingKeys.update([o[0] for o in result.fetchall()])

        inserts = []
        newKeys = []

        for newItem in itemsByKey.values():
            if newItem.key in existingKeys:
                continue

            inserts.append(dict(
                modelSetId=liveDbModelSet.id,
                key=newItem.key,
                dataType=newItem.dataType,
                rawValue=newItem.rawValue,
                displayValue=newItem.displayValue,
                importHash=newItem.importHash
            ))

            newKeys.append(newItem.key)

        if not inserts:
            return []

        conn.execute(LiveDbItem.__table__.insert(), inserts)

        transaction.commit()
        logger.info("Inserted %s LiveDbItems, %s already existed, in %s",
                    len(inserts), len(existingKeys), (datetime.now(pytz.utc) - startTime))

        return newKeys

    except Exception as e:
        transaction.rollback()
        logger.debug("Task failed, but it will retry. %s", e)
        raise self.retry(exc=e, countdown=10)

    finally:
        conn.close()
        session.close()



"""Fixed nullable Queue rawValue

Peek Plugin Database Migration Script

Revision ID: 1f014c013f8b
Revises: d4ffb9c5cb27
Create Date: 2020-04-21 18:33:51.189917

"""

# revision identifiers, used by Alembic.
revision = '1f014c013f8b'
down_revision = 'd4ffb9c5cb27'
branch_labels = None
depends_on = None

import sqlalchemy as sa
from alembic import op


def upgrade():
    schema = 'pl_livedb'
    table = 'LiveDbRawValueQueue'
    op.execute('''CREATE TEMP TABLE tempqueue AS
                  select * from %s."%s"''' % (schema, table))

    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('LiveDbRawValueQueue', schema='pl_livedb')

    op.create_table('LiveDbRawValueQueue',
                    sa.Column('id', sa.BigInteger(), autoincrement=True, nullable=False),
                    sa.Column('modelSetId', sa.Integer(), nullable=False),
                    sa.Column('key', sa.String(), nullable=False),
                    sa.Column('rawValue', sa.String(), nullable=True),
                    sa.PrimaryKeyConstraint('id'),
                    schema='pl_livedb'
                    )
    op.create_index('idx_LiveDbRawValueQueue_all', 'LiveDbRawValueQueue',
                    ['id', 'modelSetId', 'key', 'rawValue'], unique=False,
                    schema='pl_livedb')
    # ### end Alembic commands ###

    op.execute('''SELECT setval('%s."%s_id_seq"', max(id))
                  FROM %s."%s"''' % (schema, table, schema, table))

    op.execute('''INSERT INTO %s."%s"
                  select * from tempqueue''' % (schema, table))

    op.execute('''DROP TABLE tempqueue''')


def downgrade():
    pass



from peek_plugin_livedb._private.PluginNames import livedbTuplePrefix

from vortex.Tuple import Tuple, addTupleType


@addTupleType
class LiveDbRawValueTuple(Tuple):
    """ Live DB Raw Value Tuple

    This tuple represents a raw key / value pair in the Live Db

    """
    __tupleType__ = livedbTuplePrefix + 'LiveDbRawValueTuple'
    __slots__ = ("id", "key", "rawValue")

    def __init__(self, id=None, key=None, rawValue=None):
        # DON'T CALL SUPER INIT
        self.id = id
        self.key = key
        self.rawValue = rawValue



import logging
from typing import List

from twisted.internet import defer
from twisted.internet.defer import Deferred, inlineCallbacks

from peek_plugin_livedb._private.server.LiveDBReadApi import LiveDBReadApi
from peek_plugin_livedb._private.server.controller.LiveDbController import \
    LiveDbController
from peek_plugin_livedb._private.server.controller.LiveDbImportController import \
    LiveDbImportController
from peek_plugin_livedb._private.server.controller.LiveDbValueUpdateQueueController import \
    LiveDbValueUpdateQueueController
from peek_plugin_livedb.server.LiveDBWriteApiABC import LiveDBWriteApiABC
from peek_plugin_livedb.tuples.ImportLiveDbItemTuple import ImportLiveDbItemTuple
from peek_plugin_livedb.tuples.LiveDbRawValueUpdateTuple import LiveDbRawValueUpdateTuple

logger = logging.getLogger(__name__)


class LiveDBWriteApi(LiveDBWriteApiABC):

    def __init__(self):
        self._queueController = None
        self._liveDbController = None
        self._liveDbImportController = None
        self._readApi = None
        self._dbSessionCreator = None
        self._dbEngine = None

    def setup(self, queueController: LiveDbValueUpdateQueueController,
              liveDbController: LiveDbController,
              liveDbImportController: LiveDbImportController,
              readApi: LiveDBReadApi,
              dbSessionCreator,
              dbEngine):
        self._queueController = queueController
        self._liveDbController = liveDbController
        self._liveDbImportController = liveDbImportController
        self._readApi = readApi
        self._dbSessionCreator = dbSessionCreator
        self._dbEngine = dbEngine

    def shutdown(self):
        pass

    @inlineCallbacks
    def updateRawValues(self, modelSetName: str,
                        updates: List[LiveDbRawValueUpdateTuple]) -> Deferred:
        """ Update Raw Values

        """
        if not updates:
            return

        yield self._queueController.queueData(modelSetName, updates)

        self._readApi.rawValueUpdatesObservable(modelSetName).on_next(updates)

    def importLiveDbItems(self, modelSetName: str,
                          newItems: List[ImportLiveDbItemTuple]) -> Deferred:
        if not newItems:
            return defer.succeed(True)

        return self._liveDbImportController.importLiveDbItems(modelSetName, newItems)

    def prioritiseLiveDbValueAcquisition(self, modelSetName: str,
                                         liveDbKeys: List[str]) -> Deferred:
        self._readApi.priorityKeysObservable(modelSetName).on_next(liveDbKeys)
        return defer.succeed(True)

    def pollLiveDbValueAcquisition(self, modelSetName: str,
                                   liveDbKeys: List[str]) -> Deferred:
        self._readApi.pollKeysObservable(modelSetName).on_next(liveDbKeys)
        return defer.succeed(True)



from peek_plugin_livedb._private.tuples.AdminStatusTuple import \
    AdminStatusTuple
from vortex.handler.TupleDataObservableHandler import TupleDataObservableHandler

from peek_plugin_livedb._private.PluginNames import livedbFilt
from peek_plugin_livedb._private.PluginNames import livedbObservableName
from peek_plugin_livedb._private.server.controller.AdminStatusController import \
    AdminStatusController
from peek_plugin_livedb._private.server.tuple_providers.AdminStatusTupleProvider import \
    AdminStatusTupleProvider


def makeTupleDataObservableHandler(ormSessionCreator,
                                   adminStatusController: AdminStatusController):
    """" Make Tuple Data Observable Handler

    This method creates the observable object, registers the tuple providers and then
    returns it.

    :param adminStatusController:
    :param ormSessionCreator: A function that returns a SQLAlchemy session when called

    :return: An instance of :code:`TupleDataObservableHandler`

    """
    tupleObservable = TupleDataObservableHandler(
        observableName=livedbObservableName,
        additionalFilt=livedbFilt)

    # # Register TupleProviders here
    tupleObservable.addTupleProvider(AdminStatusTuple.tupleName(),
                                     AdminStatusTupleProvider(adminStatusController))
    return tupleObservable



from vortex.handler.TupleActionProcessor import TupleActionProcessor

from peek_plugin_livedb._private.PluginNames import livedbFilt
from peek_plugin_livedb._private.PluginNames import livedbActionProcessorName
from .controller.MainController import MainController


def makeTupleActionProcessorHandler(mainController: MainController):
    processor = TupleActionProcessor(
        tupleActionProcessorName=livedbActionProcessorName,
        additionalFilt=livedbFilt,
        defaultDelegate=mainController)
    return processor



from peek_plugin_livedb._private.PluginNames import livedbTuplePrefix

from vortex.Tuple import Tuple, addTupleType


@addTupleType
class LiveDbDisplayValueUpdateTuple(Tuple):
    """ Live DB Display Value Update

    This tuple represents an update to the display value for a live db item

    """
    __tupleType__ = livedbTuplePrefix + 'LiveDbDisplayValueUpdateTuple'
    __slots__ = ("key", "displayValue")



from vortex.Tuple import Tuple, addTupleType

from peek_plugin_livedb._private.PluginNames import livedbTuplePrefix
from peek_plugin_livedb._private.storage.LiveDbItem import LiveDbItem


@addTupleType
class LiveDbDisplayValueTuple(Tuple):
    """ Live DB Display Value Tuple

    This tuple stores a value of a key in the Live DB database.

    """
    __tupleType__ = livedbTuplePrefix + 'LiveDbDisplayValueTuple'
    __slots__ = ("key", "dataType", "rawValue", "displayValue")

    DATA_TYPE_NUMBER_VALUE = LiveDbItem.NUMBER_VALUE
    DATA_TYPE_STRING_VALUE = LiveDbItem.STRING_VALUE
    DATA_TYPE_COLOR = LiveDbItem.COLOR
    DATA_TYPE_LINE_WIDTH = LiveDbItem.LINE_WIDTH
    DATA_TYPE_LINE_STYLE = LiveDbItem.LINE_STYLE
    DATA_TYPE_GROUP_PTR = LiveDbItem.GROUP_PTR

    @classmethod
    def sqlCoreLoad(cls, row):
        return LiveDbDisplayValueTuple(key=row.key, dataType=row.dataType,
                                       rawValue=row.rawValue,
                                       displayValue=row.displayValue)



from peek_plugin_livedb._private.server.LiveDBReadApi import LiveDBReadApi
from peek_plugin_livedb._private.server.LiveDBWriteApi import LiveDBWriteApi
from peek_plugin_livedb._private.server.controller.LiveDbController import \
    LiveDbController
from peek_plugin_livedb._private.server.controller.LiveDbImportController import \
    LiveDbImportController
from peek_plugin_livedb._private.server.controller.LiveDbValueUpdateQueueController import \
    LiveDbValueUpdateQueueController
from peek_plugin_livedb.server.LiveDBApiABC import LiveDBApiABC
from peek_plugin_livedb.server.LiveDBReadApiABC import LiveDBReadApiABC
from peek_plugin_livedb.server.LiveDBWriteApiABC import LiveDBWriteApiABC


class LiveDBApi(LiveDBApiABC):

    def __init__(self):
        self._readApi = LiveDBReadApi()
        self._writeApi = LiveDBWriteApi()

    def setup(self, queueController: LiveDbValueUpdateQueueController,
              liveDbController: LiveDbController,
              liveDbImportController: LiveDbImportController,
              dbSessionCreator,
              dbEngine):
        self._readApi.setup(liveDbController=liveDbController,
                            dbSessionCreator=dbSessionCreator,
                            dbEngine=dbEngine)

        self._writeApi.setup(queueController=queueController,
                             liveDbController=liveDbController,
                             liveDbImportController=liveDbImportController,
                             readApi=self._readApi,
                             dbSessionCreator=dbSessionCreator,
                             dbEngine=dbEngine)

    def shutdown(self):
        self._readApi.shutdown()
        self._writeApi.shutdown()

        self._readApi = None
        self._writeApi = None

    @property
    def writeApi(self) -> LiveDBWriteApiABC:
        return self._writeApi

    @property
    def readApi(self) -> LiveDBReadApiABC:
        return self._readApi



from typing import List, Dict

from sqlalchemy import select

from peek_plugin_base.storage.StorageUtil import makeCoreValuesSubqueryCondition, \
    makeOrmValuesSubqueryCondition
from peek_plugin_livedb._private.storage.LiveDbItem import LiveDbItem
from peek_plugin_livedb._private.storage.LiveDbModelSet import LiveDbModelSet, \
    getOrCreateLiveDbModelSet
from peek_plugin_livedb.tuples.LiveDbDisplayValueTuple import LiveDbDisplayValueTuple
from peek_plugin_livedb.tuples.LiveDbRawValueTuple import LiveDbRawValueTuple


class WorkerApi:
    """ Worker Api

    This class allows other classes to work with the LiveDB plugin on the
    worker service.

    """
    _FETCH_SIZE = 5000

    @classmethod
    def getLiveDbDisplayValues(cls,
                               ormSession,
                               modelSetKey: str,
                               liveDbKeys: List[str]
                               ) -> List[LiveDbDisplayValueTuple]:
        """ Get Live DB Display Values

        Return an array of items representing the display values from the LiveDB.

        :param ormSession: The SQLAlchemy orm session from the calling code.
        :param modelSetKey: The name of the model set to get the keys for
        :param liveDbKeys: An array of LiveDb Keys.

        :returns: An array of tuples.
        """
        if not liveDbKeys:
            return []

        liveDbModelSet = getOrCreateLiveDbModelSet(ormSession, modelSetKey)

        liveDbKeys = set(liveDbKeys)  # Remove duplicates if any exist.
        qry = (
            ormSession.query(LiveDbItem)
                .filter(LiveDbItem.modelSetId == liveDbModelSet.id)
                .filter(makeOrmValuesSubqueryCondition(
                ormSession, LiveDbItem.key, list(liveDbKeys)
            ))
                .yield_per(cls._FETCH_SIZE)
        )

        results = []

        for item in qry:
            results.append(
                LiveDbDisplayValueTuple(key=item.key,
                                        displayValue=item.displayValue,
                                        rawValue=item.rawValue,
                                        dataType=item.dataType)
            )

        return results




import logging
from collections import defaultdict
from typing import List, Optional

from peek_plugin_base.storage.LoadPayloadPgUtil import getTuplesPayloadBlocking, \
    LoadPayloadTupleResult
from peek_plugin_livedb._private.server.controller.LiveDbController import \
    LiveDbController
from peek_plugin_livedb._private.storage.LiveDbItem import LiveDbItem
from peek_plugin_livedb._private.storage.LiveDbModelSet import getOrCreateLiveDbModelSet
from peek_plugin_livedb.server.LiveDBReadApiABC import LiveDBReadApiABC
from peek_plugin_livedb.tuples.LiveDbDisplayValueTuple import LiveDbDisplayValueTuple
from rx.subjects import Subject
from sqlalchemy import select
from twisted.internet.defer import Deferred
from vortex.DeferUtil import deferToThreadWrapWithLogger

logger = logging.getLogger(__name__)


class LiveDBReadApi(LiveDBReadApiABC):

    def __init__(self):
        self._liveDbController = None
        self._dbSessionCreator = None
        self._dbEngine = None

        self._prioritySubject = defaultdict(Subject)
        self._pollSubject = defaultdict(Subject)
        self._additionsSubject = defaultdict(Subject)
        self._deletionsSubject = defaultdict(Subject)
        self._rawValueUpdatesSubject = defaultdict(Subject)
        self._displayValueUpdatesSubject = defaultdict(Subject)

    def setup(self, liveDbController: LiveDbController,
              dbSessionCreator,
              dbEngine):
        self._liveDbController = liveDbController
        self._dbSessionCreator = dbSessionCreator
        self._dbEngine = dbEngine

    def shutdown(self):
        pass

    def priorityKeysObservable(self, modelSetName: str) -> Subject:
        return self._prioritySubject[modelSetName]

    def pollKeysObservable(self, modelSetName: str) -> Subject:
        return self._pollSubject[modelSetName]

    def itemAdditionsObservable(self, modelSetName: str) -> Subject:
        return self._additionsSubject[modelSetName]

    def itemDeletionsObservable(self, modelSetName: str) -> Subject:
        return self._deletionsSubject[modelSetName]

    def bulkLoadDeferredGenerator(self, modelSetName: str,
                                  keyList: Optional[List[str]] = None,
                                  chunkSize: int = 2500) -> Deferred:
        offset = 0
        limit = chunkSize
        while True:
            yield qryChunk(modelSetName, offset, limit, keyList, self._dbSessionCreator)

            offset += limit

    def rawValueUpdatesObservable(self, modelSetName: str) -> Subject:
        return self._rawValueUpdatesSubject[modelSetName]

    def displayValueUpdatesObservable(self, modelSetName: str) -> Subject:
        return self._displayValueUpdatesSubject[modelSetName]


@deferToThreadWrapWithLogger(logger)
def qryChunk(modelSetKey: str, offset: int, limit: int, keyList: List[str],
             dbSessionCreator) -> LoadPayloadTupleResult:
    # If they've given us an empty key list, that is what they will get back
    if keyList is not None and not keyList:
        return LoadPayloadTupleResult(encodedPayload=None, count=0)

    table = LiveDbItem.__table__
    cols = [table.c.key, table.c.dataType, table.c.rawValue, table.c.displayValue]

    session = dbSessionCreator()
    try:
        liveDbModelSet = getOrCreateLiveDbModelSet(session, modelSetKey)

        sql = select(cols) \
            .order_by(table.c.id) \
            .where(table.c.modelSetId == liveDbModelSet.id)

        if keyList is not None:
            sql = sql.where(table.c.key.in_(keyList))

        sql = sql.offset(offset).limit(limit)

        return getTuplesPayloadBlocking(
            dbSessionCreator,
            sql,
            LiveDbDisplayValueTuple.sqlCoreLoad,
            fetchSize=limit
        )

    finally:
        session.close()
